from typing import Any

from azure.mgmt.dns import DnsManagementClient
from sewer.dns_providers import common


class AzureDnsDns(common.BaseDns):

    def __init__(self, subscription: str, rg: str, zone: str, credential, **kwargs: Any):
        self.dns_client = DnsManagementClient(
            credential, subscription
        )
        self.rg = rg
        self.zone = zone
        super().__init__(**kwargs)

    def create_prefix(self, domain_name: str) -> str:
        prefix = "_acme-challenge" + "." + domain_name
        if domain_name.endswith(self.zone):
            zone_name_len = len(self.zone) + 1
            prefix = prefix[0:-zone_name_len]
        return prefix

    def create_dns_record(self, domain_name: str, domain_dns_value: str):
        prefix = self.create_prefix(domain_name)
        self.dns_client.record_sets.create_or_update(
            self.rg, self.zone, prefix, "TXT", {
                "ttl": 1,
                "TXTRecords": [
                    {"value": [domain_dns_value]}
                ]
            })

    def delete_dns_record(self, domain_name, domain_dns_value):
        prefix = self.create_prefix(domain_name)
        self.dns_client.record_sets.delete(self.rg, self.zone, prefix, "TXT")



import pem
import re
from typing import List, Optional
import logging

import OpenSSL
import sewer.client
from datetime import datetime, timezone

from azure.core.exceptions import ResourceNotFoundError
from azure.identity import DefaultAzureCredential
from azure.keyvault.certificates import CertificateClient, CertificatePolicy
from azure.keyvault.secrets import SecretClient
from sewer.crypto import AcmeKey, AcmeAccount

from azlet.dns_providers.azuredns import AzureDnsDns


def clean_name(domain: str) -> str:
    name = domain.lower()
    name = re.sub('[^0-9a-zA-Z]+', '-', name)
    if domain.startswith("*"):
        name = "star-" + name
    if domain.startswith("@"):
        name = "direct-" + name
    return name


class AzertBot:
    def __init__(self, keyvault_name: str, dns_subscription: str, dns_rg: str, zone: str, credential=None):
        if credential is None:
            credential = DefaultAzureCredential()
        self.certificate_client = CertificateClient(vault_url=f"https://{keyvault_name}.vault.azure.net/",
                                                    credential=credential, version="2016-10-01")
        self.secrets_client = SecretClient(vault_url=f"https://{keyvault_name}.vault.azure.net/",
                                           credential=credential)
        self.dns_class = AzureDnsDns(dns_subscription, dns_rg, zone, credential)
        self.acme_account = None

    def account(self) -> Optional[AcmeAccount]:
        if self.acme_account is None:
            try:
                secret = self.secrets_client.get_secret("acme-account-key2")
                self.acme_account = AcmeAccount.from_pem(secret.value.encode("UTF-8"))
            except ResourceNotFoundError as e:
                return None
        return self.acme_account

    def store_account(self, account: AcmeKey):
        self.secrets_client.set_secret("acme-account-key2", account.to_pem().decode('utf-8'))
        self.acme_account = account

    def store_pfx(self, domain: str, cert: str, key: AcmeKey, tags=None, name=None):
        if not name:
            name = clean_name(domain)

        all_certs = [OpenSSL.crypto.load_certificate(OpenSSL.crypto.FILETYPE_PEM, pem.as_bytes()) for pem in
                     pem.parse(cert.encode('utf-8'))]
        key_pem = OpenSSL.crypto.load_privatekey(OpenSSL.crypto.FILETYPE_PEM, key.to_pem())
        pkcs = OpenSSL.crypto.PKCS12()
        pkcs.set_privatekey(key_pem)
        pkcs.set_certificate(all_certs[0])
        pkcs.set_ca_certificates(all_certs[1:])
        pfx_cert = pkcs.export()
        if tags:
            self.certificate_client.import_certificate(name, pfx_cert, tags=tags)
        else:
            self.certificate_client.import_certificate(name, pfx_cert)

    def create_certificate(self, domain_names: List[str], tags=None, secret_name=None):
        account = self.account()
        account_is_new = False
        if account is None:
            account = AcmeAccount.create("rsa3072")
            account_is_new = True
        client = sewer.client.Client(
            domain_name=domain_names[0],
            domain_alt_names=domain_names[1:],
            provider=self.dns_class,
            account=account,
            is_new_acct=account_is_new,
            cert_key=AcmeKey.create("rsa3072"),
            LOG_LEVEL="WARNING"
        )
        if account_is_new:
            client.acme_register()
            self.store_account(account)

        cert = client.get_certificate()
        key = client.cert_key

        self.store_pfx(domain_names[0], cert, key, tags, name=secret_name)

    def check_exists(self, domain_name):
        name = clean_name(domain_name)
        try:
            cert = self.certificate_client.get_certificate(name)
            raise Warning(
                "Cannot create certificate: certificate for this name is already stored. Use 'rotate' to renew the certificate, or use --force-creation.")
        except ResourceNotFoundError:
            pass

    def create(self, prefix: List[str], force=False, tags=None):
        domain_names = [ p + "." + self.dns_class.zone if p != "@" else self.dns_class.zone for p in prefix ]
        if not force:
            self.check_exists(domain_names[0])
        self.create_certificate(domain_names, tags)
        

    def rotate(self, threshold: int = 14):
        for props in self.certificate_client.list_properties_of_certificates():
            if not props.enabled:
                continue
            logging.info(f"Checking {props.name}")
            diff = props.expires_on - datetime.now(timezone.utc)
            logging.info(f"Expires in {diff.days}d. (Threshold: {threshold}d)")
            if diff.days > threshold:
                continue
            cert = self.certificate_client.get_certificate(props.name)
            domain_name = cert.policy.subject.replace("CN=", "")
            subject_alt_names = [ san for san in (cert.policy.san_dns_names or []) if san != domain_name ]
            logging.info(f'Certificate {props.name} has subject {domain_name} and SANs {subject_alt_names}')
            logging.info(f'Checking subject: "{domain_name}" should end with "{self.dns_class.zone}"')
            if not str(domain_name).endswith(self.dns_class.zone):
                continue
            logging.info(f"Starting renewal ...")
            self.create_certificate(domain_names=[domain_name]+subject_alt_names, tags=props.tags, secret_name=props.name)

    def rotate_domain(self, prefix: str):
        domain_name = prefix + '.' + self.dns_class.zone
        name = clean_name(domain_name)
        try:
            cert = self.certificate_client.get_certificate(name)
            # is None or dict of tags
            tags = cert.properties.tags
            self.create_certificate(domain_names=[domain_name], tags=tags)
        except:
            logging.error("Cannot find certificate to renew.")



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='azlet',
    version='0.9.1',
    packages=['azlet', 'azlet.dns_providers'],
    url='https://github.com/claasd/azlet',
    license='MIT',
    author='Claas Diederichs',
    author_email='',
    description='Python package to create/renew certificates using azure DNS and azure KeyVault',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        "pem",
        "sewer==0.8.4",
        "azure-keyvault-certificates~=4.2",
        "azure-keyvault-secrets~=4.2",
        "azure-identity~=1.5",
        "azure-mgmt-dns~=8.0",
        "pyopenssl"
    ]
)



import argparse
import logging

from azure.identity import DefaultAzureCredential

from azlet.azertbot import AzertBot
logging.basicConfig(level=logging.INFO)
logging.getLogger('azure').setLevel(logging.WARN)

def execute():
    parser = argparse.ArgumentParser()
    parser.add_argument("operation", choices=["create", "rotate"])
    parser.add_argument("--keyvault-name", "-k", help="name of the keyvault")
    parser.add_argument("--dns-zone", "-d", help="name of the dns-zone")
    parser.add_argument("--dns-subscription", "-s", help="subscription of the dns zone")
    parser.add_argument("--dns-resource-group", "-g", help="resource-group of the dns zone")
    parser.add_argument("--prefix", "-p", help="name of the certificate prefix (only for create operation) Use '@' to create a certificate for the root domain. Can be supplied multiple times.",
                        required=False, action='append' )
    parser.add_argument("--tags", "-t", nargs="*", action = keyvalue, help="tags of the certificate", default=False)
    parser.add_argument("--force-creation",
                        action='store_true',
                        help="Try to create a new certificate even if a certificate already exists in the key vault",
                        required=False)
    parser.add_argument("--rotation-threshold",
                        type=int,
                        default=30,
                        help="Number of days a rotation should be done before certificate expiration.")
    parser.add_argument("--exclude-cli-credential", type=bool, default=False,
                        help="Whether to exclude the Azure CLI from the credential.")
    parser.add_argument("--exclude-environment-credential", type=bool, default=False,
                        help="Whether to exclude a service principal configured by environment variables from the credential.")
    parser.add_argument("--exclude-managed-identity-credential", type=bool, default=False,
                        help="Whether to exclude managed identity from the credential.")
    parser.add_argument("--exclude-visual-studio-code-credential", type=bool, default=False,
                        help="Whether to exclude stored credential from VS Code.")
    parser.add_argument("--exclude-shared-token-cache-credential", type=bool, default=False,
                        help="Whether to exclude the shared token cache.")
    parser.add_argument("--exclude-interactive-browser-credential", type=bool, default=True,
                        help="Whether to exclude interactive browser authentication")
    args = parser.parse_args()

    credential = DefaultAzureCredential(
        exclude_cli_credential=args.exclude_cli_credential,
        exclude_environment_credential=args.exclude_environment_credential,
        exclude_managed_identity_credential=args.exclude_managed_identity_credential,
        exclude_visual_studio_code_credential=args.exclude_visual_studio_code_credential,
        exclude_shared_token_cache_credential=args.exclude_shared_token_cache_credential,
        exclude_interactive_browser_credential=args.exclude_interactive_browser_credential
    )
    bot = AzertBot(keyvault_name=args.keyvault_name, dns_subscription=args.dns_subscription,
                   dns_rg=args.dns_resource_group, zone=args.dns_zone, credential=credential)
    if args.operation == "rotate":
        bot.rotate(args.rotation_threshold)

    if args.operation == "create":
        if not args.prefix:
            parser.error("--prefix is required for command 'create'")
        if args.tags:
            bot.create(args.prefix, args.force_creation, args.tags)
        else:
            bot.create(args.prefix, args.force_creation)


class keyvalue(argparse.Action):
    def __call__( self , parser, namespace, values, option_string = None):
        setattr(namespace, self.dest, dict())
          
        for value in values:
            # split it into key and value
            key, value = value.split('=')
            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


if __name__ == "__main__":
    execute()
